#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # 2A.ml - Réduction d'une forêt aléatoire - correction
# 
# Le modèle Lasso permet de sélectionner des variables, une forêt aléatoire produit une prédiction comme étant la moyenne d'arbres de régression. Et si on mélangeait les deux ?

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# In[2]:


get_ipython().run_line_magic('matplotlib', 'inline')


# ## Datasets
# 
# Comme il faut toujours des données, on prend ce jeu [Diabetes](https://scikit-learn.org/stable/modules/generated/sklearn.datasets.load_diabetes.html).

# In[3]:


from sklearn.datasets import load_diabetes
data = load_diabetes()
X, y = data.data, data.target


# In[4]:


from sklearn.model_selection import train_test_split
X_train, X_test, y_train, y_test = train_test_split(X, y)


# ## Une forêt aléatoire

# In[5]:


from sklearn.ensemble import RandomForestRegressor as model_class
clr = model_class()
clr.fit(X_train, y_train)


# Le nombre d'arbres est...

# In[6]:


len(clr.estimators_)


# In[7]:


from sklearn.metrics import r2_score
r2_score(y_test, clr.predict(X_test))


# ## Random Forest = moyenne des prédictions
# 
# On recommence en faisant la moyenne soi-même.

# In[8]:


import numpy
dest = numpy.zeros((X_test.shape[0], len(clr.estimators_)))
estimators = numpy.array(clr.estimators_).ravel()
for i, est in enumerate(estimators):
    pred = est.predict(X_test)
    dest[:, i] = pred

average = numpy.mean(dest, axis=1)
r2_score(y_test, average)


# A priori, c'est la même chose.

# ## Pondérer les arbres à l'aide d'une régression linéaire
# 
# La forêt aléatoire est une façon de créer de nouvelles features, 100 exactement qu'on utilise pour caler une régression linéaire.

# In[9]:


from sklearn.linear_model import LinearRegression


def new_features(forest, X):
    dest = numpy.zeros((X.shape[0], len(forest.estimators_)))
    estimators = numpy.array(forest.estimators_).ravel()
    for i, est in enumerate(estimators):
        pred = est.predict(X)
        dest[:, i] = pred
    return dest


X_train_2 = new_features(clr, X_train)
lr = LinearRegression()
lr.fit(X_train_2, y_train)


# In[10]:


X_test_2 = new_features(clr, X_test)
r2_score(y_test, lr.predict(X_test_2))


# Un peu moins bien, un peu mieux, le risque d'overfitting est un peu plus grand avec ces nombreuses features car la base d'apprentissage ne contient que 379 observations (regardez ``X_train.shape`` pour vérifier).

# In[11]:


lr.coef_


# In[12]:


import matplotlib.pyplot as plt
fig, ax = plt.subplots(1, 1, figsize=(12, 4))
ax.bar(numpy.arange(0, len(lr.coef_)), lr.coef_)
ax.set_title("Coefficients pour chaque arbre calculés avec une régression linéaire");


# Le score est avec une régression linéaire sur les variables initiales est nettement moins élevé.

# In[13]:


lr_raw = LinearRegression()
lr_raw.fit(X_train, y_train)
r2_score(y_test, lr_raw.predict(X_test))


# ## Sélection d'arbres
# 
# L'idée est d'utiliser un algorithme de sélection de variables type [Lasso](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Lasso.html) pour réduire la forêt aléatoire sans perdre en performance. C'est presque le même code.

# In[14]:


from sklearn.linear_model import Lasso

lrs = Lasso(max_iter=10000)
lrs.fit(X_train_2, y_train)
lrs.coef_


# Pas mal de zéros donc pas mal d'arbres non utilisés.

# In[15]:


r2_score(y_test, lrs.predict(X_test_2))


# Pas trop de perte... Ca donne envie d'essayer plusieurs valeur de `alpha`.

# In[16]:


from tqdm import tqdm
alphas = [0.01 * i for i in range(100)] +[1 + 0.1 * i for i in range(100)]
obs = []
for i in tqdm(range(0, len(alphas))):
    alpha = alphas[i]
    lrs = Lasso(max_iter=20000, alpha=alpha)
    lrs.fit(X_train_2, y_train)
    obs.append(dict(
        alpha=alpha,
        null=len(lrs.coef_[lrs.coef_!=0]),
        r2=r2_score(y_test, lrs.predict(X_test_2))
    ))


# In[17]:


from pandas import DataFrame
df = DataFrame(obs)
df.tail()


# In[18]:


fig, ax = plt.subplots(1, 2, figsize=(12, 4))
df[["alpha", "null"]].set_index("alpha").plot(ax=ax[0], logx=True)
ax[0].set_title("Nombre de coefficients non nulls")
df[["alpha", "r2"]].set_index("alpha").plot(ax=ax[1], logx=True)
ax[1].set_title("r2");


# Dans ce cas, supprimer des arbres augmente la performance, comme évoqué ci-dessus, cela réduit l'overfitting. Le nombre d'arbres peut être réduit des deux tiers avec ce modèle.

# In[19]: